/* This file is public domain. */

#include "image.h"

#include <png.h>
#include <stdio.h>
#include <stdlib.h>

#include "os_dep.h"

/* loads the image name and put it to image structure.
 * returns -1 if error, 0 otherwise.
 * This function reads only 8b paletted images. I don't need anything else.
 */
int load_image(image *im, char *name)
{
  FILE *f=fopen(name, "rb");
  char header[8];
  png_structp png_ptr;
  png_infop info_ptr;
  png_infop end_info;
  png_bytep *rowptrs;
  png_colorp pal;
  png_uint_32 width, height;
  int bit_depth, color_type;
  int i;

  /* checking the png file */
  if (!f) {
    perror("read_image.fopen");
    return -1;
  }
  if (fread(header, 1, 8, f)!=8) {
    perror("read_image : trying to read the header");
    fclose(f);
    return -1;
  }
  if (!png_check_sig(header, 8)) {
    fprintf(stderr, "The file \"%s\" is not a png image\n", name);
    fclose(f);
    return -1;
  }
  /* creating structures for reading */
  if (!(png_ptr= png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)0, 0, 0))) {
    fprintf(stderr, "Error 1 with png loader\n");
    fclose(f);
    return -1;
  }
  if (!(info_ptr= png_create_info_struct(png_ptr))) {
    fprintf(stderr, "Error 2 with png loader\n");
    png_destroy_read_struct(&png_ptr, (png_infopp)0, (png_infopp)0);
    fclose(f);
    return -1;
  }
  if (!(end_info= png_create_info_struct(png_ptr))) {
    fprintf(stderr, "Error 3 with png loader\n");
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)0);
    return -1;
  }
  /* initing IO */
  png_init_io(png_ptr, f);
  png_set_sig_bytes(png_ptr, 8);

  /* ok, now read it */
  /* first, the info stuff */
  png_read_info(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, (int *)0, (int *)0, (int *)0);
#ifdef DEBUG
  fprintf(stderr, "file=\"%s\" width=%ld height=%ld bit depth=%d\n", name, width, height, bit_depth);
  switch(color_type) {
    case PNG_COLOR_TYPE_GRAY :
      fprintf(stderr, "PNG_COLOR_TYPE_GRAY\n");
      break;
    case PNG_COLOR_TYPE_GRAY_ALPHA :
      fprintf (stderr, "PNG_COLOR_TYPE_GRAY_ALPHA\n");
      break;
    case PNG_COLOR_TYPE_PALETTE :
      fprintf(stderr, "PNG_COLOR_TYPE_PALETTE\n");
      break;
    case PNG_COLOR_TYPE_RGB :
      fprintf(stderr, "PNG_COLOR_TYPE_RGB\n");
      break;
    case PNG_COLOR_TYPE_RGB_ALPHA :
      fprintf(stderr, "PNG_COLOR_TYPE_RGB_ALPHA\n");
      break;
    default :
      fprintf(stderr, "%d : unkown format\n", color_type);
  }
#endif
  if (bit_depth!=8) {
    fprintf(stderr, "bad image type (not 8b), impossible to read\n");
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  if (color_type!=PNG_COLOR_TYPE_PALETTE) {
    fprintf(stderr, "bad image type (not paletted), impossible to read\n");
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  /* get the palette */
  if (png_get_PLTE(png_ptr, info_ptr, &pal, &im->palsize)==0) {
    fprintf(stderr, "read_image error with loading the palette\n");
    exit(1);
  }
  if (im->palsize>256) {
    fprintf(stderr, "palette size (%d) too big !\n", im->palsize);
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  if (!(im->palette=(char *)malloc(im->palsize*3))) {
    perror("read_image error with mallocing memory for the palette");
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  for (i=0; i<im->palsize; i++) {
    im->palette[i*3+0]=pal[i].red;
    im->palette[i*3+1]=pal[i].green;
    im->palette[i*3+2]=pal[i].blue;
  }
#ifdef DEBUG
  fprintf(stderr, "palette size=%d\n", im->palsize);
#endif
  /* now, read the image */
  /* get some memory */
  if (!(im->data=(char *)malloc(width*height))) {
    perror("read_image.malloc 1");
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  if (!(rowptrs=(png_bytep *)malloc(height*sizeof(png_bytep)))) {
    perror("read_image.malloc 2");
    free(im->data);
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(f);
    return -1;
  }
  for (i=0; i<height; i++)
    rowptrs[i]=(png_bytep)((unsigned int)(im->data)+i*width);
  /* load the image */
  png_read_image(png_ptr, rowptrs);
  png_read_end(png_ptr, end_info);
  /* loading ok, free what is not needed more */
  im->width=width;
  im->height=height;
  free(rowptrs);
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
  fclose(f);
  return 0;
}

/* no check is done to see if it fits, no clipping */
void put_image(image *im, int x, int y)
{
  char *p=ecran1+y*320+x, *q=im->data;
  int i;
  for (i=0; i<im->height; i++, p+=320, q+=im->width)
    memcpy(p, q, im->width);
}

void free_image(image *im)
{
  free(im->data);
  free(im->palette);
}
